# Traffic Light System

## Problem Description

### Summary

This test is a take home test, please return once completed 

### Requirements

In this test we would like you to implement a traffic light system. We are required to have 4 sets of lights, as follows. 

- Lights 1: Traffic is travelling south 
- Lights 2: Traffic is travelling west 
- Lights 3: Traffic is travelling north
- Lights 4: Traffic is travelling east

The lights in which traffic is travelling on the same axis can be green at the same time. During normal hours all lights stay green for 20 seconds, but during peak times north and south lights are green for 40 seconds while west and east are green for 10 seconds. Peak hours are 0800 to 1000 and 1700 to 1900. Yellow lights are shown for 5 seconds before red lights are shown. Red lights stay on until the cross-traffic is red for at least 4 seconds, once a red light goes off then the green is shown for the required time(eg the sequence is reset). 

Bonus: At this intersection north bound traffic has a green right-turn signal, which stops the south bound traffic and allows north bound traffic to turn right. This is green at the end of north/south green light and stays green for 10 seconds. During this time north bound is green, north right-turn is green and all other lights are red. 

### Implementation/Outcomes

1.	Implement a front-end and backend (you can use ‘dotnet new’ templates of your choice)
2.	The backend will contain the logic and state of the running traffic lights. The front-end will be a visual representation of the traffic lights, with the data served from the backend. 
3.	There’s no need to have a perfect design on the front end, something simple and functional is fine (unless this is an area of strength you would like to show off). Noting* we will review the client side code.
4.	There’s no need to implement entity framework (or similar) to store the data in a database, a in-memory store is fine
5.	Code needs to follow architecture & best practices for enterprise grade systems

Note: Code will be evaluated not just for function, but on the quality of the code. 

# Design

## Algorithmic Assumptions

The problem clearly specifies a kind of timed "simulation" with a decoupled front- and back-ends.  The problem has been approached on that basis.

Additionally:

- The algorithm uses the term `period` to denote specific timing profiles (e.g. off-peak vs peak period).  Within each `period` is a single implicit `sequence` that itself is comprised of a set of timed `states` that each fully specify the colour of each light.  These are stored in JSON format in the file `periods.json`.
- The problem description does not specify behaviour on period change (i.e. transitioning from normal timing to period timing).  One would assume that an instantaneous change would easily create unpredictable and unsafe conditions for any hypothetical drivers, and so in this implementation the current full `sequence` is allowed to complete before transitioning to the next period.  For this example, a full cycle should not last longer than approximately 80 seconds and therefore this is the expected maximum delay between period transitions.

Period data is configurable/stored in `periods.json`.

## Environment and Infrastructure

As per the problem description, C# ASP.NET has been used. A web front-end was chosen due to the broad use of HTML/CSS/JavaScript and the ease with which it can be changed or updated according to need.  It would also lend itself particularly well to a scenario where the front- and back-ends would exist on different machines.

The server uses an IHostedService `TrafficLightSimulatorService` to simulate the traffic light changes.  By default, this service updates four times per simulated second.

The UI is a simple web page that polls for updates from the server.  Due to the use of locks to avoid race conditions, this can induced minor delays (milliseconds at worse) in the processing and updates of traffic light information in the simulation thread.  The main loop is designed in such a way that this will not cause cumulative delays in traffic light timing.  However, it is worth considering that another approach may be desirable (such as pushed client updates rather than polling) if, for instance, many clients were expected to be making simultaneous requests of the server.

Incorporating some feedback on an earlier version of this solution in Python/Django, the UI now only polls when it knows the server is due for a state update.  The update thread continues to run, however, ensuring that the displayed simulation time is continuously updated.

The main frontend web page, including the JavaScript polling code, is specified in its entirety in `Pages/Index.cshtml`. **In a production project, CSS, JavaScript, and HTML would all be separated into individual packages**.  However for ease of review these remain inline in a single file in this project.

# Installation and Usage

The project was written in Visual Studio Community 2022 (17.7.6) and targets the .NET 6.0 framework.

## Execution

The project should run directly from Visual Studio.  It has not been tested for deployment.

## Configuration

Settings are in the standard `appsettings.json`.  The following keys are used:

1. **PeriodConfigurationFile**: The file containining the period/state data for traffic light behaviour.
1. **StartTime**: The internal time at which the simulation starts, in ISO text format.  By default this is 07:59 for demonstration purposes (as it will show the changeover from morning off-peak to morning peak).
1. **TimeFactor**: The time "acceleration" factor (floating point) for the simulation.  A factor of 1.0 will run in real-time, and higher factors will run faster.  Default is 4.0.

